#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import copy
from random import shuffle

from thrift.protocol import TCompactProtocol, TBinaryProtocol
from thrift.transport import TSocket
from thrift.transport import TTransport

from flumelogger.errors import (
    ConfigurationError,
    ConnectionFailure,
    ServerSelectionError,
    OperationFailure)
from flumelogger.flumeng import ThriftSourceProtocol
from flumelogger.flumeog import ThriftFlumeEventServer
from flumelogger.utils import log_debug
from flumelogger.utils import split_hosts


class FlumeEventServer(object):
    def __init__(self, host="localhost", port=9090, timeout=1000, type='ng', reuse=True, debug=False):
        """ Init the event server stuff (nodes and co).

        Args:
            host (Optional[str or list]): hostname or IP address, string or a list of hostname.
            port (Optional[int]): port number on which to connect.
            timeout (Optional[int]): Controls how long the socket will wait for a response.
            type (Optional[str]): The flume agent generation.
            reuse (bool): using a single TCP connection, otherwhise open/close connection for each operation.
            debug (bool): True show debug messages, otherwhise is quiet.
        """
        self.host = host
        if isinstance(self.host, basestring):
            self.host = [self.host]
        self.port = port
        self.timeout = timeout
        self.type = type
        self.reuse = reuse
        self.debug = debug

        # NOTE: parse the host URI.
        for entity in self.host:
            self.default_nodes = split_hosts(entity, self.port)
            log_debug("nodes available {0}".format(self.default_nodes), debug=self.debug)

        if not self.default_nodes:
            raise ConfigurationError("need to specify at least one host")

        # NOTE: init stuff to manage the pool.
        self.active_nodes = [node for node in self.default_nodes]
        self.open_connections = {}

    def _connect(self, host, port):
        """ Connect to the thrift source.

        Args:
            host (str): hostname or IP address.
            port (int): port number on which to connect.

        Returns:
            client, transport: The client/transport from thrift.

        Raises:
            ConnectionFailure: connection to the thrift source cannot be made os is lost.
        """
        try:
            socket = TSocket.TSocket(host, port)
            socket._timeout = self.timeout
            if self.type == 'ng':
                transport = TTransport.TFramedTransport(socket)
                protocol = TCompactProtocol.TCompactProtocol(transport)
                client = ThriftSourceProtocol.Client(protocol)
            elif self.type == 'og':
                transport = TTransport.TBufferedTransport(socket)
                protocol = TBinaryProtocol.TBinaryProtocol(transport)
                client = ThriftFlumeEventServer.Client(protocol)
            transport.open()
        except Exception as e:
            raise ConnectionFailure(e)
        else:
            return client, transport

    def _is_connected(self, node):
        """ Check if a TCP connection is already opened.

        Args:
            host (str): Hostname/port combo.

        Returns:
            bool: True if an active connection is available, False otherwise.
        """
        return node in self.open_connections and self.open_connections[node]['transport'].isOpen()

    def _remove_node(self, node):
        """ Remove a node from the active nodes pool.

        Args:
            host (str): Hostname/port combo.
        """
        log_debug("remove {0} from the active nodes".format(node), debug=self.debug)
        self.open_connections.pop(node, None)
        if node in self.active_nodes:
            self.active_nodes.remove(node)

    def _add_node(self, node, client, transport):
        """ Add a node to the active nodes pool.

        Args:
            host (str): Hostname/port combo.
        """
        log_debug("add {0} to the active nodes".format(node), debug=self.debug)
        self.open_connections[node] = {'client': client, 'transport': transport}
        if node not in self.active_nodes:
            self.active_nodes.append(node)

    def __enter__(self):
        """ Connect to an active node.
        """
        try:
            if not self.active_nodes:
                raise StopIteration

            shuffle(self.active_nodes)
            for node in copy(self.active_nodes):
                try:
                    if not self.active_nodes:
                        raise StopIteration

                    # NOTE: try to re-use a connection.
                    self.current_node = node
                    if self.reuse and self._is_connected(node=self.current_node):
                        log_debug('re-using connection to {0}'.format(self.current_node), debug=self.debug)
                        return self.open_connections[self.current_node]['client']

                    # NOTE: new connection.
                    self.client, self.transport = self._connect(*self.current_node)
                    self._add_node(node=self.current_node, client=self.client, transport=self.transport)
                    log_debug('opened new connection to {0}'.format(self.current_node), debug=self.debug)
                except ConnectionFailure as e:
                    self._remove_node(node=self.current_node)
                else:
                    return self.client
        except StopIteration:
            raise ServerSelectionError("no server available")

    def __exit__(self, type, value, unpack):
        """ Ensure to close the transport connection after each operation.
        """
        if not self.reuse and hasattr(self, "transport"):
            self.transport.close()

    def reconnect(self):
        """ Helper will attempt to reconnect on nodes.
        """
        for node in self.default_nodes:
            try:
                if node not in self.active_nodes:
                    log_debug('reconnecting to {0}'.format(node), debug=self.debug)
                    client, transport = self._connect(*node)
                    self._add_node(node=node, client=client, transport=transport)
            except ConnectionFailure as e:
                log_debug('failed to reconnect to {0}'.format(node), debug=self.debug)

    def append(self, event, client):
        try:
            return client.append(event)
        except Exception as e:
            self._remove_node(node=self.current_node)
            raise OperationFailure(e)

    def append_batch(self, events, client):
        try:
            return client.appendBatch(events)
        except Exception as e:
            self._remove_node(node=self.current_node)
            raise OperationFailure(e)



# -*- coding: utf-8 -*-
import re

from flumelogger.errors import ConfigurationError


def parse_host(entity, default_port):
    """@Author: pymongo driver

    Validates a host string

    Returns a 2-tuple of host followed by port where port is default_port
    if it wasn't specified in the string.

    :Parameters:
        - `entity`: A host or host:port string where host could be a
                    hostname or IP address.
        - `default_port`: The port number to use when one wasn't
                          specified in entity.
    """
    host = entity
    port = default_port
    if entity.find(':') != -1:
        if entity.count(':') > 1:
            raise ValueError("Reserved characters such as ':' must be "
                             "escaped according RFC 2396. An IPv6 "
                             "address literal must be enclosed in '[' "
                             "and ']' according to RFC 2732.")
        host, port = host.split(':', 1)
    if isinstance(port, str):
        if not port.isdigit() or int(port) > 65535 or int(port) <= 0:
            raise ValueError("Port must be an integer between 0 and 65535: %s"
                             % (port,))
        port = int(port)

    # Normalize hostname to lowercase, since DNS is case-insensitive:
    # http://tools.ietf.org/html/rfc4343
    # This prevents useless rediscovery if "foo.com" is in the node list but
    # "FOO.com" is in the ismaster response.
    return host.lower(), port

def split_hosts(hosts, default_port):
    """@Author: pymongo driver

    takes a string of the form host1[:port],host2[:port]... and
    splits it into (host, port) tuples. if [:port] isn't present the
    default_port is used.

    returns a set of 2-tuples containing the host name (or ip) followed by
    port number.

    :parameters:
        - `hosts`: a string of the form host1[:port],host2[:port],...
        - `default_port`: the port number to use when one wasn't specified
          for a host.
    """
    nodes = []
    for entity in hosts.split(','):
        if not entity:
            raise ConfigurationError("empty host "
                "(or extra comma in host list).")
        port = default_port
        nodes.append(parse_host(entity, port))
    return nodes

def log_debug(msg, debug):
    """ Simple helper to display message to avoid logging module.

    Args:
        msg (str): the message to display.
        debug (bool): True display message, False otherwise.
    """
    if debug:
        print msg



class ConfigurationError(Exception):
    """Raised when something is incorrectly configured."""


class ConnectionFailure(Exception):
    """Raised when a connection to the thrift source cannot be made os is lost."""


class ServerSelectionError(Exception):
    """Thrown when no thrift source is available for an operation."""


class OperationFailure(Exception):
    """Raised when a thrift operation fails."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ast
import logging
import socket
import time

from datetime import datetime

from flumelogger.eventserver import FlumeEventServer
from flumelogger.flumeng.ttypes import ThriftFlumeEvent as ThriftFlumeNGEvent
from flumelogger.flumeog.ttypes import ThriftFlumeEvent as ThriftFlumeOGEvent

PRIORITY = {"FATAL": 0,
            "CRITICAL": 0,
            "ERROR": 1,
            "WARNING": 2,
            "INFO": 3,
            "DEBUG": 4,
            "TRACE": 5}


class FlumeHandler(logging.Handler):
    def __init__(self, host="localhost", port=9090, timeout=1000, type='ng',
                 reuse=True, debug=False, headers=None):
        if headers is None:
            headers = {}

        # run the regular Handler __init__
        logging.Handler.__init__(self)

        # default header detection
        if 'host' not in headers:
            headers['host'] = socket.gethostname()

        self.host = host
        self.port = port
        self.timeout = timeout
        self.type = type
        self.reuse = reuse
        self.debug = debug
        self.headers = headers
        self.eventserver = FlumeEventServer(host=self.host,
                                            port=self.port,
                                            timeout=self.timeout,
                                            type=self.type,
                                            reuse=self.reuse,
                                            debug=self.debug)
        self.reconnect = self.eventserver.reconnect

    def event_ng(self, body, headers):
        return ThriftFlumeNGEvent(headers=headers, body=body)

    def event_og(self, body, fields):
        pri = PRIORITY[fields['pri']]
        dt = int(time.time() * 1000)
        ns = datetime.now().microsecond * 1000
        host = fields['host']

        del fields['pri']
        del fields['host']

        return ThriftFlumeOGEvent(timestamp=dt,
                                  priority=pri,
                                  body=body,
                                  host=host,
                                  nanos=ns,
                                  fields=fields)

    def parse_record(self, record):
        body = self.format(record)
        headers = self.headers.copy()
        try:
            msg = ast.literal_eval(body)
        except Exception:
            msg = None
        if isinstance(msg, dict):
            if 'message' in msg:
                body = msg['message']
                del msg['message']
            else:
                body = ""
            headers.update(msg)
        headers['pri'] = record.levelname.upper()
        return body, headers

    def emit(self, record):
        if isinstance(record.msg, list):
            self.emit_many(record)
        else:
            self.emit_one(record)

    def emit_one(self, record):
        try:
            event = {'ng': self.event_ng, 'og': self.event_og}
            body, headers = self.parse_record(record)
            try:
                tevent = event[self.type](body, headers)
            except KeyError:
                raise Exception('Wrong flume type specified')

            # send event
            with self.eventserver as client:
                result = self.eventserver.append(tevent, client)
                if result != 0:
                    raise Exception('Thrift error {}'.format(result))
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            record.msg = e
            self.handleError(record)

    def emit_many(self, records):
        try:
            events = []
            for obj in records.msg:
                record = logging.LogRecord(name=records.name,
                                           level=records.levelno,
                                           pathname=records.pathname,
                                           lineno=records.lineno,
                                           msg=obj,
                                           args=records.args,
                                           exc_info=records.exc_info)
                event = {'ng': self.event_ng, 'og': self.event_og}
                body, headers = self.parse_record(record)
                try:
                    tevent = event[self.type](body, headers)
                except KeyError:
                    raise Exception('Wrong flume type specified')
                else:
                    events.append(tevent)

            # send events
            with self.eventserver as client:
                result = self.eventserver.append_batch(events, client)
                if result != 0:
                    raise Exception('Thrift error {}'.format(result))
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            record = logging.LogRecord(name=records.name,
                                       level=records.levelno,
                                       pathname=records.pathname,
                                       lineno=records.lineno,
                                       msg=e,
                                       args=records.args,
                                       exc_info=records.exc_info)
            self.handleError(record)
